﻿
using System;
using System.Collections.Generic;
using Autodesk.Revit.DB;
using Autodesk.Revit.UI;

namespace StairsAutomation
{
	/// <summary>
	/// Description of StairsBoundedLandingConfiguration.
	/// </summary>
	public class StairsBoundedLandingConfiguration : IStairsLandingConfiguration
	{
		public StairsBoundedLandingConfiguration(Line stairsRunBoundary1, Line stairsRunBoundary2, XYZ runDirection, double elevation, IList<Curve> boundaries,
		                                        Autodesk.Revit.Creation.Application appCreate)
		{
			// offset to base elevation
            m_stairsRunBoundary1 = LandingConfigurationUtils.ProjectCurveToElevation(stairsRunBoundary1, elevation) as Line;
            m_stairsRunBoundary2 = LandingConfigurationUtils.ProjectCurveToElevation(stairsRunBoundary2, elevation) as Line;
            
            m_runDirection = runDirection;
            m_boundaries = LandingConfigurationUtils.ProjectCurvesToElevation(boundaries, elevation);
            m_elevation = elevation;
            m_appCreate = appCreate;
		}
		
		private Line m_stairsRunBoundary1;
        private Line m_stairsRunBoundary2;
        private XYZ m_runDirection;
        private IList<Curve> m_boundaries;
        private double m_elevation;
        private Autodesk.Revit.Creation.Application m_appCreate;
		
		public Autodesk.Revit.DB.CurveLoop GetLandingBoundary()
		{
			// Project line to boundary
			Line line = LandingConfigurationUtils.FindLongestEndpointConnection(m_stairsRunBoundary1, m_stairsRunBoundary2);
			Line unboundLine = Line.get_Unbound(line.get_EndPoint(0), line.get_EndPoint(1) - line.get_EndPoint(0));
			
			// Find intersections
			int intersectionCurveIndex1 = -1;
			double intersectionParameter1 = -1.0;
			int intersectionCurveIndex2 = -1;
			double intersectionParameter2 = -1.0;
			
			
			for (int  i = 0; i < m_boundaries.Count; i++ )
			{
				Curve curve = m_boundaries[i];
				IntersectionResultArray resultArray;
				SetComparisonResult result = curve.Intersect(unboundLine, out resultArray);
				if (result == SetComparisonResult.Overlap)
				{
					if (resultArray.Size > 2)
					{
						throw new Exception("Too many intersections.");
					}
					
					for (int j = 0; j < resultArray.Size; j++)
					{
						IntersectionResult intersection = resultArray.get_Item(j);
						if (intersectionCurveIndex1 < 0)
						{
							intersectionCurveIndex1 = i;
							XYZ point = intersection.XYZPoint;
							intersectionParameter1 = LandingConfigurationUtils.ComputeParameter(curve, point);
						}
						else if (intersectionCurveIndex2 < 0)
						{
							intersectionCurveIndex2 = i;
							XYZ point = intersection.XYZPoint;
							intersectionParameter2 = LandingConfigurationUtils.ComputeParameter(curve, point);
						}	
						else
						{
							throw new Exception("Too many intersections.");
						}
					}				
				}
			}
			
			if (intersectionCurveIndex1 < 0 || intersectionCurveIndex2 < 0)
				throw new Exception("Not enough intersections.");
			
			// project direction to find one curve in target loop
			XYZ lineMidpoint = line.Evaluate(0.5, true);
			Line directionLine = Line.get_Unbound(lineMidpoint, m_runDirection);
			int targetCurveIndex = -1;
			
			for (int  i = 0; i < m_boundaries.Count; i++ )
			{
				Curve curve = m_boundaries[i];
				IntersectionResultArray resultArray;
				SetComparisonResult result = curve.Intersect(directionLine, out resultArray);
				if (result == SetComparisonResult.Overlap)
				{
					if (resultArray.Size > 2)
					{
						throw new Exception("Too many intersections.");
					}
					
					for (int j = 0; j < resultArray.Size; j++)
					{
						IntersectionResult intersection = resultArray.get_Item(j);
						XYZ point = intersection.XYZPoint;
						XYZ direction = point - lineMidpoint;
						double angle = direction.AngleTo(m_runDirection);
						if (Math.Abs(angle) < 1e-06)
						{
							targetCurveIndex = i;
							break;
						}
					}				
				}
				if (targetCurveIndex >= 0)
					break;
			}
			
			if (targetCurveIndex < 0)
			{
				throw new Exception("Could not determine projected side.");
			}
			
			
			
			// Create curve loop, starting with target curve
			List<Curve> curvesForCurveLoop = new List<Curve>();
			if (intersectionCurveIndex1 == targetCurveIndex || intersectionCurveIndex2 == targetCurveIndex)
			{
				throw new Exception("TODO need to split target curve!");
			}
			
			
			int startIndex = -1;
			int endIndex = -1;
			
			// All wrong - use CurveChainIntersection
			if (intersectionCurveIndex1 < targetCurveIndex && intersectionCurveIndex2 < targetCurveIndex)
			{
				startIndex = intersectionCurveIndex1 < intersectionCurveIndex2 ? intersectionCurveIndex2 : intersectionCurveIndex1;
				endIndex = startIndex == intersectionCurveIndex2 ? intersectionCurveIndex1 : intersectionCurveIndex2;
			}
			else if (targetCurveIndex < intersectionCurveIndex1 && targetCurveIndex < intersectionCurveIndex2)
			{
				startIndex = intersectionCurveIndex1 < intersectionCurveIndex2 ? intersectionCurveIndex2 : intersectionCurveIndex1;
				endIndex = startIndex == intersectionCurveIndex2 ? intersectionCurveIndex1 : intersectionCurveIndex2;
			}
			else
			{
				startIndex = intersectionCurveIndex1 < intersectionCurveIndex2 ? intersectionCurveIndex1 : intersectionCurveIndex2;
				endIndex = startIndex == intersectionCurveIndex2 ? intersectionCurveIndex1 : intersectionCurveIndex2;
			}
			
			
			// Cut start index curve.
			// Start at intersection parameter to end
			Curve startCurve = m_boundaries[startIndex];
			double endParameter = startCurve.get_EndParameter(1);
			Curve trimmedCurve = GetTrimmedCurve(startCurve, startIndex == intersectionCurveIndex1 ? intersectionParameter1 : intersectionParameter2, endParameter);
			curvesForCurveLoop.Add(trimmedCurve);
			
			// Copy middle curves
			if (startIndex <= endIndex)
			{
				for (int i = startIndex + 1; i < endIndex; i++)
				{
					curvesForCurveLoop.Add(m_boundaries[i].Clone());
				}
			}
			else
			{
				for (int i = startIndex + 1; i < m_boundaries.Count; i++)
				{
					curvesForCurveLoop.Add(m_boundaries[i].Clone());
				}
				for (int i = 0; i < endIndex -1; i++)
				{
					curvesForCurveLoop.Add(m_boundaries[i].Clone());
				}
			}
			
			// Cut end index curve.
			Curve endCurve = m_boundaries[endIndex];
			double startParameter = startCurve.get_EndParameter(0);
			trimmedCurve = GetTrimmedCurve(endCurve, startParameter, endIndex == intersectionCurveIndex1 ? intersectionParameter1 : intersectionParameter2);
			curvesForCurveLoop.Add(trimmedCurve);
			
			// Add line at stair run ends
			// Should be from end intersection point to start intersection point
			
			
			curvesForCurveLoop.Add(Line.get_Bound(m_boundaries[endIndex].Evaluate(endIndex == intersectionCurveIndex1 ? intersectionParameter1 : intersectionParameter2, false),
			                                      m_boundaries[startIndex].Evaluate(startIndex == intersectionCurveIndex1 ? intersectionParameter1 : intersectionParameter2, false)));
			
			/*
			foreach (Curve curve in curvesForCurveLoop)
			{
				TaskDialog.Show("Curve", String.Format("Start {0} end {1}", curve.get_EndPoint(0), curve.get_EndPoint(1)));
			}
			*/
			
			return CurveLoop.Create(curvesForCurveLoop);
		}
		
		private Curve GetTrimmedCurve(Curve curve, double startParameter, double endParameter)
		{
			XYZ start = curve.Evaluate(startParameter, false);
			XYZ end = curve.Evaluate(endParameter, false);
			if (curve is Line)
			{	
				return Line.get_Bound(start, end);
			}
			else if (curve is Arc)
			{
				// Get 3rd point
				double midParameter = (startParameter + endParameter) / 2.0;
				XYZ mid = curve.Evaluate(midParameter, false);
				
				return m_appCreate.NewArc(start, end, mid);
			}
			else throw new Exception("Unsupported curve type for get trimmed curve.");
		}
		
		private class CurveChainIntersection
		{
			public int Index { get; set; }
			public double Param { get; set; }
		}
		
		public double GetLandingBaseElevation()
		{
			return m_elevation;
		}
	}
}
