﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Autodesk.Revit.DB;
using Autodesk.Revit.DB.Architecture;
using Autodesk.Revit.UI;

namespace StairsAutomation
{
    class StairsAutomationUtility
    {
        private Autodesk.Revit.DB.Document document;
        private IList<Reference> curveReferences;
        private Stairs m_stairs;
        private int m_stairsNumber;

        private StairsAutomationUtility(Autodesk.Revit.DB.Document document, int stairsNumber)
        {
            this.document = document;
            this.m_stairsNumber = stairsNumber;
        }

        private static Tuple<Level, Level, Level> FindTargetLevels(Document doc, String name1, String name2, String name3)
        {
            FilteredElementCollector collector = new FilteredElementCollector(doc);
            collector.OfClass(typeof(Level));

            Level level1 = null;
            Level level2 = null;
            Level level3 = null;
            foreach (Level level in collector.Cast<Level>().Where<Level>(level => level.Name.Equals(name1) || level.Name.Equals(name2) || level.Name.Equals(name3)))
            {
                if (level.Name.Equals(name1))
                    level1 = level;
                else if (level.Name.Equals(name2))
                    level2 = level;
                else
                    level3 = level;
            }

            Tuple<Level, Level, Level> ret = new Tuple<Level, Level, Level>(level1, level2, level3);

            return ret;
        }

        public static StairsAutomationUtility Create(Autodesk.Revit.DB.Document document, int stairsNumber)
        {
            StairsAutomationUtility utility = new StairsAutomationUtility(document, stairsNumber);
            return utility;
        }

        private void SetupLevels()
        {
            Tuple<Level, Level, Level> targetLevels = FindTargetLevels(document, "Level 1", "Level 2", "Level 3");
            Level level1 = targetLevels.Item1;
            Level level2 = targetLevels.Item2;
            Level level3 = targetLevels.Item3;
            switch (m_stairsNumber)
            {
                // Straight run 1 level
                case 0:
                    BottomLevel = level1;
                    TopLevel = level2;
                    break;
                
                // Standard stair 1 level
                case 1:
                        BottomLevel = level1;
                        TopLevel = level2;
                        break;
                // Standard stair multi-level
                case 2:
                        BottomLevel = level1;
                        TopLevel = level3;
                        break;
                // Curved run 1 level
                case 3:
                        BottomLevel = level1;
                        TopLevel = level2;
                        break;
                // Curve run 2 level
                case 4:
                        BottomLevel = level1;
                        TopLevel = level3;
                        break;

                // Boundary curves
                case 5:
                        BottomLevel = level1;
                        TopLevel = level3;
                        break;

            }
        }

        private IStairsConfiguration GetConfiguration()
        {
            switch (m_stairsNumber)
            {
                    // Straight run 1 level
                case 0:
                    return  new StairsSingleStraightRun(m_stairs, BottomLevel, Transform.get_Translation(new XYZ(30, 0, 0)));

                // Standard stair 1 level
                case 1:
                    {
                        StairsStandardRun configuration = new StairsStandardRun(m_stairs, BottomLevel, 1, Transform.get_Rotation(new XYZ(-20, -20, 0), XYZ.BasisZ, Math.PI/4.0));
                        configuration.EqualizeRuns = true;
                        configuration.Initialize();
                        return configuration;
                    }
                    // Standard stair multi-level
                case 2:
                    {
                     StairsStandardRun configuration = new StairsStandardRun(m_stairs, BottomLevel, 3, Transform.get_Rotation(new XYZ(15, -25, 0), XYZ.BasisZ, 7.0 * Math.PI / 6.0));
                        configuration.RunWidth = 6.0;
                        configuration.RunOffset = 8.0 / 12.0;
                        configuration.LandingWidth = 4.0;
                        configuration.EqualizeRuns = true;
                        configuration.Initialize();
                        return configuration;
                    }
                // Curved run 1 level
                case 3:
                    return new StairsSingleCurvedRun(m_stairs, BottomLevel, 6.0, Transform.get_Rotation(new XYZ(-20, 30, 0), XYZ.BasisZ, Math.PI / 2.0));
                // Curve run 2 level
                case 4:
                    return new StairsSingleCurvedRun(m_stairs, BottomLevel, 4.0, Transform.get_Rotation(new XYZ(-10, -20, 0), XYZ.BasisZ, Math.PI));

                case 5:
                    {
                        StairsBoundRun configuration = new StairsBoundRun(m_stairs, BottomLevel, 3);
                        configuration.RunWidth = 6.0;
                        //configuration.RunOffset = 8.0 / 12.0;
                        configuration.EqualizeRuns = true;
                        configuration.SetupRunOffset();
                        configuration.Initialize();
                        return configuration;
                    }

            }
            return null;
        }

        internal void SetCurves(IList<Autodesk.Revit.DB.Reference> references)
        {
            throw new NotImplementedException();
        }

        public Autodesk.Revit.DB.Level BottomLevel { get; set; }

        public Autodesk.Revit.DB.Level TopLevel { get; set; }

        private IStairsConfiguration GetConfigurationOld()
        {
            //List<IStairsConfiguration> ret = new List<IStairsConfiguration>();

            //ret.Add(new StairsTransformedRunConfiguration(new 
            //StairsSingleStraightRun configuration =  new StairsSingleStraightRun(m_stairs, BottomLevel, Transform.get_Rotation(new XYZ(5, 10, 0), XYZ.BasisZ, Math.PI / 6.0));
            //StairsSingleCurvedRun configuration = new StairsSingleCurvedRun(m_stairs, BottomLevel, 4.0, Transform.get_Rotation(new XYZ(5, 10, 0), XYZ.BasisZ, Math.PI / 6.0));
            StairsStandardRun configuration = new StairsStandardRun(m_stairs, BottomLevel, 3, Transform.get_Rotation(new XYZ(5, 10, 0), XYZ.BasisZ, Math.PI / 6.0));
            configuration.RunWidth = 6.0;
            configuration.RunOffset = 8.0 / 12.0;
            configuration.LandingWidth = 4.0;
            configuration.EqualizeRuns = true;
            configuration.Initialize();
            //StairsStandardRun configuration = new StairsStandardRun(m_stairs, BottomLevel, 1, Transform.Identity);
            //StairsTransformedRunConfiguration configuration = new StairsTransformedRunConfiguration(baseConfiguration, );
            //configuration.BottomLevel = BottomLevel;
            //configuration.TopLevel = TopLevel;
            return configuration;
        }

        internal void GenerateStairs()
        {
            using (StairsEditScope editScope = new StairsEditScope(document, "Stairs Automation"))
            {
                SetupLevels();

                ElementId stairsElementId = editScope.Start(BottomLevel.Id, TopLevel.Id);
                m_stairs = document.GetElement(stairsElementId) as Stairs;

                Transaction t = new Transaction(document, "Stairs Automation");
                t.Start();

                IStairsConfiguration configuration = GetConfiguration();
                if (configuration == null)
                    return;

                int numberOfRuns = configuration.GetNumberOfRuns();
                for (int i = 0; i < numberOfRuns; i++)
                {
                    StairsRun.CreateSketchedRun(document, stairsElementId, configuration.GetRunElevation(i), 
                        configuration.GetRunBoundaryCurves(i), configuration.GetRunRiserCurves(i), configuration.GetRunStairsPath(i));
                }

                int numberOfLandings = configuration.GetNumberOfLandings();
                for (int i = 0; i < numberOfLandings; i++)
                {
                    StairsLanding.CreateSketchedLanding(document, stairsElementId,
                        configuration.GetLandingBoundary(i), configuration.GetLandingBaseElevation(i));
                }

                t.Commit();

                editScope.Commit();
            }
        }

        private static bool IsStoryLevel(Level level)
        {
            Parameter p = level.get_Parameter(BuiltInParameter.LEVEL_IS_BUILDING_STORY);
            return (p != null && p.AsInteger() != 0);
        }

        private bool IsBetweenExtents(Level level)
        {
            if (level.Id == BottomLevel.Id || level.Id == TopLevel.Id)
                return false;
            return level.Elevation > BottomLevel.Elevation && level.Elevation < TopLevel.Elevation;
        }

        private IEnumerable<Level> FindStoryLevelsBetweenExtents()
        {
            FilteredElementCollector collector = new FilteredElementCollector(document);
            collector.OfClass(typeof(Level));

            return collector.Cast<Level>().Where<Level>(level => IsBetweenExtents(level) && IsStoryLevel(level));
        }
    }
}
