﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Autodesk.Revit;
using Autodesk.Revit.DB;
using Autodesk.Revit.DB.IFC;
using Autodesk.Revit.DB.ExtensibleStorage;

namespace BIM.IFC.Export.UI
{
    public class IFCExportConfigurationsMap
    {
        private Dictionary<String, IFCExportConfiguration> m_configurations = new Dictionary<String, IFCExportConfiguration>();
        private Schema m_schema = null;
        private static Guid s_schemaId = new Guid("A1E672E5-AC88-4933-A019-F9068402CFA7");

        public IFCExportConfigurationsMap()
        {
        }

        public IFCExportConfigurationsMap(IFCExportConfigurationsMap map)
        {
            // Deep copy
            foreach (IFCExportConfiguration value in map.Values)
            {
                Add(value.Clone());
            }
        }

        public void AddBuiltInConfigurations()
        {
            // These are the built-in configurations.  Provide a more extensible means of storage.
            // Order of construction: name, version, space boundaries, QTO, split walls, internal sets, 2d elems.
            Add(IFCExportConfiguration.CreateBuiltInConfiguration("Default 2x3", IFCVersion.IFC2x3, 1, false, false, true, true));
            Add(IFCExportConfiguration.CreateBuiltInConfiguration("Default 2x2", IFCVersion.IFC2x2, 1, false, false, true, false));
            Add(IFCExportConfiguration.CreateBuiltInConfiguration("BCA", IFCVersion.IFCBCA, 1, false, true, true, false));
            Add(IFCExportConfiguration.CreateBuiltInConfiguration("GSA", IFCVersion.IFCCOBIE, 2, true, true, true, true));
            Add(IFCExportConfiguration.CreateBuiltInConfiguration("Certification", IFCVersion.IFC2x3CV2, 0, false, true, false, false));
        }

        public void AddSavedConfigurations(Document document)
        {
            if (m_schema == null)
            {
                m_schema = Schema.Lookup(s_schemaId);
            }

            if (m_schema != null)
            {
                foreach (DataStorage storedSetup in GetSavedConfigurations(document))
                {
                    Entity configEntity = storedSetup.GetEntity(m_schema);
                    IFCExportConfiguration configuration = IFCExportConfiguration.CreateDefaultConfiguration();
                    configuration.Name = configEntity.Get<String>(s_setupName);
                    //configuration.Description = configEntity.Get<String>(s_setupDescription);
                    configuration.IFCVersion = (IFCVersion)configEntity.Get<int>(s_setupVersion);
                    configuration.IFCFileType = (IFCFileFormat)configEntity.Get<int>(s_setupFileFormat);
                    configuration.SpaceBoundaries = configEntity.Get<int>(s_setupSpaceBoundaries);
                    configuration.ExportBaseQuantities = configEntity.Get<bool>(s_setupQTO);
                    configuration.SplitWallsAndColumns = configEntity.Get<bool>(s_splitWallsAndColumns);
                    configuration.Export2DElements = configEntity.Get<bool>(s_setupExport2D);
                    configuration.ExportInternalRevitPropertySets = configEntity.Get<bool>(s_setupExportRevitProps);
                    configuration.Use2DRoomBoundaryForVolume = configEntity.Get<bool>(s_setupUse2DForRoomVolume);
                    configuration.UseFamilyAndTypeNameForReference = configEntity.Get<bool>(s_setupUseFamilyAndTypeName);

                    Add(configuration);
                } 
            }   
        }

        private const String s_setupName = "Name";
        private const String s_setupDescription = "Description";
        private const String s_setupVersion = "Version";
        private const String s_setupFileFormat = "FileFormat";
        private const String s_setupSpaceBoundaries = "SpaceBoundaryLevel";
        private const String s_setupQTO = "ExportBaseQuantities";
        private const String s_splitWallsAndColumns = "SplitWallsAndColumns";
        private const String s_setupCurrentView = "VisibleElementsInCurrentView";
        private const String s_setupExport2D = "Export2DElements";
        private const String s_setupExportRevitProps = "ExportInternalRevitPropertySets";
        private const String s_setupUse2DForRoomVolume = "Use2DBoundariesForRoomVolume";
        private const String s_setupUseFamilyAndTypeName = "UseFamilyAndTypeNameForReference";

        public void UpdateSavedConfigurations(Document document)
        {
            if (m_schema == null)
            {
                m_schema = Schema.Lookup(s_schemaId);
            }

            // Are there any setups to save or resave?
            List<IFCExportConfiguration> setupsToSave = new List<IFCExportConfiguration>();
            foreach (IFCExportConfiguration configuration in m_configurations.Values)
            {
                if (configuration.IsBuiltIn)
                    continue;

                // Store in-session settings in the cached in-session configuration
                if (configuration.IsInSession)
                {
                    IFCExportConfiguration.SetInSession(configuration);
                    continue;
                }

                setupsToSave.Add(configuration);
           }

           // If there are no setups to save, and if the schema is not present (which means there are no
           // previously existing setups which might have been deleted) we can skip the rest of this method.
           if (setupsToSave.Count <= 0 && m_schema == null)
               return;

           if (m_schema == null)
           {
                SchemaBuilder builder = new SchemaBuilder(s_schemaId);
                builder.SetSchemaName("IFCExportConfiguration");
                builder.AddSimpleField(s_setupName, typeof(String));
                builder.AddSimpleField(s_setupDescription, typeof(String));
                builder.AddSimpleField(s_setupVersion, typeof(int));
                builder.AddSimpleField(s_setupFileFormat, typeof(int));
                builder.AddSimpleField(s_setupSpaceBoundaries, typeof(int));
                builder.AddSimpleField(s_setupQTO, typeof(bool));
                builder.AddSimpleField(s_splitWallsAndColumns, typeof(bool));
                builder.AddSimpleField(s_setupCurrentView, typeof(bool));
                builder.AddSimpleField(s_setupExport2D, typeof(bool));
                builder.AddSimpleField(s_setupExportRevitProps, typeof(bool));
                builder.AddSimpleField(s_setupUse2DForRoomVolume, typeof(bool));
                builder.AddSimpleField(s_setupUseFamilyAndTypeName, typeof(bool));

                m_schema = builder.Finish();
           }

           // Overwrite all saved configs with the new list
           Transaction transaction = new Transaction(document, "Update IFC export setups");
           transaction.Start();
           IList<DataStorage> savedConfigurations = GetSavedConfigurations(document);
           int savedConfigurationCount = savedConfigurations.Count<DataStorage>();
           int savedConfigurationIndex = 0;
           foreach (IFCExportConfiguration configuration in setupsToSave)
           {
                DataStorage configStorage;
                if (savedConfigurationIndex >= savedConfigurationCount)
                {
                    configStorage = DataStorage.Create(document);
                }
                else
                {
                    configStorage = savedConfigurations[savedConfigurationIndex];
                    savedConfigurationIndex ++;
                }
                Entity entity = new Entity(m_schema);
                entity.Set(s_setupName, configuration.Name);
                entity.Set(s_setupDescription, configuration.Description);
                entity.Set(s_setupVersion, (int)configuration.IFCVersion);
                entity.Set(s_setupFileFormat, (int)configuration.IFCFileType);
                entity.Set(s_setupSpaceBoundaries, configuration.SpaceBoundaries);
                entity.Set(s_setupQTO, configuration.ExportBaseQuantities);
                entity.Set(s_setupCurrentView, configuration.VisibleElementsOfCurrentView);
                entity.Set(s_splitWallsAndColumns, configuration.SplitWallsAndColumns);
                entity.Set(s_setupExport2D, configuration.Export2DElements);
                entity.Set(s_setupExportRevitProps, configuration.ExportInternalRevitPropertySets);
                entity.Set(s_setupUse2DForRoomVolume, configuration.Use2DRoomBoundaryForVolume);
                entity.Set(s_setupUseFamilyAndTypeName, configuration.UseFamilyAndTypeNameForReference);

                configStorage.SetEntity(entity);
            }

            List<ElementId> elementsToDelete = new List<ElementId>();
            for (; savedConfigurationIndex < savedConfigurationCount; savedConfigurationIndex++)
            {
                DataStorage configStorage = savedConfigurations[savedConfigurationIndex];
                elementsToDelete.Add(configStorage.Id);
            }
            if (elementsToDelete.Count > 0)
                document.Delete(elementsToDelete);

            transaction.Commit();
        }

        private IList<DataStorage> GetSavedConfigurations(Document document)
        {
            FilteredElementCollector collector = new FilteredElementCollector(document);
            collector.OfClass(typeof(DataStorage));
            Func<DataStorage, bool> hasTargetData = ds => ds.GetEntity(m_schema) != null;

            return collector.Cast<DataStorage>().Where<DataStorage>(hasTargetData).ToList<DataStorage>();
        }

        public void Add(IFCExportConfiguration configuration)
        {
            m_configurations.Add(configuration.Name, configuration);
        }

        public void Remove(String name)
        {
            m_configurations.Remove(name);
        }

        public bool HasName(String name)
        {
            if (name == null) return false;
            return m_configurations.ContainsKey(name);
        }

        public IFCExportConfiguration this[String name]
        {
            get
            {
                return m_configurations[name];
            }
        }

        public IEnumerable<IFCExportConfiguration> Values
        {
            get
            {
                return m_configurations.Values;
            }
        }
    }
}
